from datetime import datetime, timedelta, timezone
import logging


class MonkeeRedis:
    def __init__(
        self, cfName, redisClient, projectName, userUid, humanDocRef, inDebugMode=False
    ):

        self.callingCF = cfName
        self.redis = redisClient
        self.userUid = userUid
        self.projectName = projectName
        self.humanDocRef = humanDocRef
        self.inDebugMode = inDebugMode
        self.sessionId = None
        self.timestamp = datetime.now(timezone.utc)
        self.minsBetweenSessions = 30
        self.timeSinceLastEvent = None

    def dprint(self, stringg):
        if self.inDebugMode:
            print(stringg)

    def set_last_cf_call(self):
        """Sets the last time that this CF was called in this project by this particular user"""
        fieldName = "CFcall={}".format(self.callingCF)
        self.set_project_human_val(
            fieldName=fieldName, dataType="datetime", value=self.timestamp
        )

    def get_last_cf_call(self):
        """Gets the last time that this CF was called in this project by this particular user"""

        fieldName = "CFcall={}".format(self.callingCF)
        print("get_last_cf_call starting for {}".format(fieldName))
        return self.get_project_human_val(fieldName=fieldName)

    def get_sec_since_last_cf_call(self):
        llt = self.get_last_cf_call()

        if llt is None:
            print("get_sec_since_last_cf_call ={}".format("None"))
            return 1000
        else:
            timeDiff = self.timestamp - llt
            print("get_sec_since_last_cf_call ={}".format(
                timeDiff.total_seconds()))
            return timeDiff.total_seconds()

    def set_project_human_val(self, fieldName, dataType, value, expireInSeconds=None):
        """Sets the compound key [self.projectName + ":" + self.userUid + ":" + fieldName] to
        value [dataType + "|" + str(value)]
        """
        key = self.projectName + ":" + self.userUid + ":" + fieldName
        val = dataType + "|" + str(value)
        self.dprint("set_project_human_val: key={}, val={}".format(key, val))
        if self.redis is not None:
            self.redis.set(key, val)
            if expireInSeconds is not None:
                self.redis.expire(key, expireInSeconds)

    def get_project_human_val(self, fieldName):
        """Gets the compound key [self.projectName + ":" + self.userUid + ":" + fieldName]
        from redis and casts it back to native based on the [dataType] used in its initial storing
        """
        if self.redis is not None:
            key = self.projectName + ":" + self.userUid + ":" + fieldName
            val = self.redis.get(key)
            if val is None or val == "None":
                return None
            else:
                val = val.decode("utf-8")
            splitted = val.split("|")
            self.dprint("get_project_human_val splitted: {} ".format(splitted))
            dataType = splitted[0]
            value = splitted[1]
            self.dprint(
                "get_project_human_val: {} = {} ({})".format(
                    key, value, dataType)
            )
            return self.format_val(dataType, value)
        else:
            return None

    def format_val(self, dataType, val):
        """Formats the [val] value given the [dataType]"""
        if val is None:
            return None
        try:
            if dataType == "datetime":
                return datetime.strptime(val, "%Y-%m-%d %H:%M:%S.%f%z")
            elif dataType == "int":
                return int(val)
            else:
                return val
        except ValueError as e:
            logging.error(repr(e))
            return None

    def get_session_id(self):
        self.sessionId = self.get_project_human_val("sessionId")
        print("get_session_id={}".format(self.sessionId))
        return self.sessionId

    def calc_session_id(self, time_diff):
        self.get_session_id()
        if self.sessionId is None:
            self.sessionId = 1
            self.set_session_id()
            return self.sessionId, None
        elif time_diff > self.minsBetweenSessions * 60:
            self.sessionId += 1
            print("incrementing session ID to {}".format(self.sessionId))
            self.set_session_id()
            self.timeSinceLastEvent = time_diff
            return self.sessionId, self.timeSinceLastEvent
        elif time_diff <= self.minsBetweenSessions * 60:
            print("keeping session ID as {}".format(self.sessionId))
            return self.sessionId, self.timeSinceLastEvent

    def set_session_id(self):
        self.set_project_human_val("sessionId", "int", self.sessionId)

        self.humanDocRef.set(
            {
                "sessionNumber": self.sessionId,
                "session": {
                    "tag": "session" + str(self.sessionId),
                    "answeredAt": self.timestamp,
                },
            },
            merge=True,
        )



from serpentmonkee import SecretMonkee, NeoMonkee

from neo4j import GraphDatabase, basic_auth  # , ServiceUnavailable
from neo4j.exceptions import ServiceUnavailable
import firebase_admin
from firebase_admin import credentials, firestore
import logging
from google.cloud import secretmanager
import os
import time
from datetime import datetime, timedelta, timezone
import json
from google.cloud.firestore_v1.transforms import DELETE_FIELD
from serpentmonkee import SecretMonkee, NeoMonkee
import sqlalchemy
import pg8000
import UtilsMonkee as um
import redis

from CypherTransaction import CypherTransactionBlock, CypherTransactionBlockWorker
from CypherQueue import CypherQueue, CypherQueues

# pip3 install neo4j==4.1.1
# pip3 install neo4j==1.7.6

if __name__ == '__main__':

    project_id = "monkee-int"  # <-- THE LOCAL DEV ENVIRONMENT

    if project_id == "monkee-prod":
        credfp = "/Users/lwicverhoef/gcp_jsons/monkee-prod-firebase-adminsdk-xvdxy-95ddc9c905.json"
    elif project_id == "monkee-dev":
        credfp = "/Users/lwicverhoef/gcp_jsons/monkee-dev-firebase-adminsdk-qw55p-b132b8a850.json"
        sqldb = sqlalchemy.create_engine(
            "postgresql+pg8000://postgres:pingping@localhost:1234/postgres")
    elif project_id == "monkee-int":
        credfp = "/Users/lwicverhoef/gcp_jsons/monkee-int.json"
        sqldb = sqlalchemy.create_engine(
            "postgresql+pg8000://postgres:g3i92lfowE8c1ED7@localhost:1236/postgres"
        )

    cred = credentials.Certificate(credfp)
    os.environ["GOOGLE_APPLICATION_CREDENTIALS"] = credfp

    secretclient = secretmanager.SecretManagerServiceClient()

    firebase_admin.initialize_app(
        cred,
        {
            "projectId": project_id,
        },
    )

    sm = SecretMonkee(secretclient, project_id, ['neo']).getSecrets()

    neomnkee = NeoMonkee(neoDriver=None,
                         sqlClient=sqldb,
                         callingCF="testingCF")

    neomnkee.makeNeoDriver(sm['neo_uri'], sm['neo_user'], sm['neo_pass'])

    def test_readWrite():
        query = """MATCH (h:humans {_project:$projectname})
        RETURN h.uid as uid, h.stepNumber as step limit 10
        """
        params = {'projectname': 'Sandbox'}
        res = neomnkee.readResults(query=query,
                                   params=params,
                                   cfInstanceUid='1234')

        params = {'projectname': 'Sandbox', 'uid': '0000'}
        query = """MATCH (h:humans {uid:$uid, _project:$projectname})
        SET h.val1 = 'one'
        RETURN h.uid as uid, h.stepNumber as step
        """

        res = neomnkee.writeResults(query=query,
                                    params=params,
                                    cfInstanceUid='1234')

        batch = [
            '01602cb23de544e8b33c4612810e96a5',
            '016a8a2f414c43b49b656b655da07fbe',
            '01f62dc44f6d4e85b0c2a7a973f97750'
        ]
        query = """
                    UNWIND $batch AS hB
                    MERGE (h:humans:test { uid: hB, _project: $projectname })
                    set h.val = 314159
                    return h.uid as uid
                    """

        params = {'projectname': 'Sandbox'}
        res = neomnkee.writeResultsBatch(query=query,
                                         params=params,
                                         batch=batch,
                                         cfInstanceUid='1234')

        for r in res:
            print(r['uid'])

    redisClient = redis.Redis()
    cQH = CypherQueue("cypherQ_High")
    cQM = CypherQueue("cypherQ_Medium")
    cQL = CypherQueue("cypherQ_Low")
    wQ = CypherQueue("cypherWorking")
    compQ = CypherQueue("cypherDone")
    queues = [cQH, cQM, cQL]

    cqs = CypherQueues(redisClient=redisClient,
                       cQueues=queues, workingQ=wQ, completedQ=compQ)

    for i in range(20):
        statements = [
            {
                "cypher":
                """MERGE (c:interactionTemplates:test {uid: $contentUid, _project:$projectName})-[:TEST]->(t:tag:test {_project:$projectName, uid:"t1"})<-[:TEST2]-(h:humans:test {_project:$projectName, uid:"AAA"})
            RETURN id(c)""",
                "parameters": {
                    "contentUid": um.get_uuid(),
                    "projectName": "testing"
                }
            },
            {
                "cypher":
                """MERGE (c:interactionTemplates:test {uid: $contentUid, _project:$projectName})-[:TEST]->(t:tag:test {_project:$projectName, uid:"t2"})<-[:TEST2]-(h:humans:test {_project:$projectName, uid:"AAA"})
            RETURN id(c)""",
                "parameters": {
                    "contentUid": um.get_uuid(),
                    "projectName": "testing"
                }
            },
        ]
        ctb = CypherTransactionBlock(priority='M',
                                     statements=statements,
                                     transactionUid=um.get_uuid(),
                                     origin='TEST')

        cqs.pushCtbToWaitingQ(ctb)

    cqs.getQLens()

    w = CypherTransactionBlockWorker(neoMonkee=neomnkee, cypherQueues=cqs)
    w.goToWork(forHowLong=100)
    cqs.getQLens()



# _METADATA_:Version: 1
# _METADATA_:Timestamp: 2020-10-01 11:59:18.409826+00:00
# _METADATA_:
class GCPDocument:
    def __init__(self, data, context):
        self.data = data
        self.context = context

        path_parts = context.resource.split("/documents/")[1].split("/")
        self.path_parts = path_parts
        self.documentUid = path_parts[-1]
        self.collectionPath = "/".join(path_parts[:-1])
        self.collectionName = path_parts[-2]
        self.projectName = path_parts[1]
        self.humanUid = None

        self.collection_path = path_parts[0]
        self.document_path = "/".join(path_parts[1:])
        self.whatChanged = []

        self.operation = "<unknown>"
        if data["oldValue"] == {} and data["value"] != {}:
            self.operation = "create"
        elif data["oldValue"] != {} and data["value"] == {}:
            self.operation = "delete"
        elif data["updateMask"] != {}:
            self.operation = "update"
            if "fieldPaths" in data["updateMask"]:
                self.whatChanged = data["updateMask"]["fieldPaths"]

        self.dataValue = self.getval(data["value"], "fields", {})
        self.dataOldValue = self.getval(data["oldValue"], "fields", {})

        self.dictValue = {}
        self.dictOldValue = {}

    def getval(self, dictionary, key, default_value=None):
        if dictionary is not None:
            if key in dictionary:
                ret = dictionary[key]
            else:
                ret = default_value
        else:
            ret = default_value
        return ret

    def decide_on_doc_change(self, ignoreIfChangeIsNotIn):
        """Decides if the doc change is worth a graph update.
        If any of the items in whatChanged is in ignoreIfChangeIsNotIn AND the operation is an UPDATE then YEs, else NO
        """
        if self.operation not in ("update"):
            return False

        for wc in self.whatChanged:
            if wc in ignoreIfChangeIsNotIn:
                return True

        return False

    def translateVal(self, fieldName, flatDict):
        if isinstance(flatDict, dict):
            for k in flatDict:
                dataType = k
                val = flatDict[k]
        else:
            dataType = fieldName
            val = flatDict
        if dataType == "doubleValue":
            return float(val)
        elif dataType == "stringValue":
            return str(val)
        elif dataType == "timestampValue":
            return str(val)  # 2020-09-07T14:42:07.635247Z
        elif dataType == "integerValue":
            return int(val)
        elif dataType == "booleanValue":
            return val

    def translateDict2(self, dict_):
        retDict = {}
        for dataType in dict_:
            # print("translateDict2. k={}, type={}".format(dataType, type(dict_[dataType])))
            if dataType not in ("arrayValue", "mapValue"):
                return self.translateVal(dataType, dict_[dataType])
            elif dataType in ("arrayValue"):
                lst = []
                for entry in dict_["arrayValue"]["values"]:
                    lst.append(self.translateDict(entry))
                retDict[dataType] = lst
            elif dataType in ("mapValue"):
                map = {}
                for field in dict_["mapValue"]["fields"]:
                    map[field] = self.translateDict2(
                        dict_["mapValue"]["fields"][field])
                return map
        return retDict

    def translateDict(self, dict_):
        # see https://cloud.google.com/functions/docs/calling/cloud-firestore
        retDict = {}
        if len(dict_) > 0:
            for k in dict_:
                # print("translateDict. k={}, type={}".format(k, type(dict_[k])))
                for k2 in dict_[k]:
                    dataType = k2
                if dataType not in ("arrayValue", "mapValue"):
                    retDict[k] = self.translateVal(k, dict_[k])
                elif dataType in ("arrayValue"):
                    lst = []
                    for entry in dict_[k]["arrayValue"]["values"]:
                        lst.append(self.translateDict2(entry))
                    retDict[k] = lst
                elif dataType in ("mapValue"):
                    map = {}
                    for field in dict_[k]["mapValue"]["fields"]:
                        map[field] = self.translateDict2(
                            dict_[k]["mapValue"]["fields"][field]
                        )
                    retDict[k] = map
        return retDict

    def change_or_removed(self, old_dict, new_dict):
        removed_keys = []
        changed_keys = []
        added_keys = []
        for k in old_dict:
            if k not in new_dict:
                removed_keys.append({k: old_dict[k]})
            else:
                if old_dict[k] != new_dict[k]:
                    changed_keys.append(
                        {k: {"oldVal": old_dict[k], "val": new_dict[k]}}
                    )

        for k in new_dict:
            if k not in old_dict:
                added_keys.append({k: new_dict[k]})

        return removed_keys, changed_keys, added_keys




import requests
from dateutil import parser
import json
from datetime import datetime, timezone
import time
import sys
import random
import uuid
import copy


# --------------------------------------------------------------------


class RoundTripEncoder(json.JSONEncoder):
    DATE_FORMAT = "%Y-%m-%d"
    TIME_FORMAT = "%H:%M:%S.%f"

    def default(self, obj):
        if isinstance(obj, datetime):
            return {
                "_type": "datetime",
                "value": obj.strftime("%s %s" % (self.DATE_FORMAT, self.TIME_FORMAT)),
            }
        return super(RoundTripEncoder, self).default(obj)


class RoundTripDecoder(json.JSONDecoder):
    def __init__(self, *args, **kwargs):
        json.JSONDecoder.__init__(
            self, object_hook=self.object_hook, *args, **kwargs)

    def object_hook(self, obj):
        if "_type" not in obj:
            return obj
        type = obj["_type"]
        if type == "datetime":
            return parser.parse(obj["value"])
        return obj


def call_cloud_function(url, headers, data):
    """
    Calls the cloud function at url with headers and data passed as part of the POST.
    Returns JSON response, passed through RoundTripDecoder
    """

    response_data = None

    try:
        response = requests.post(url=url, data=data, headers=headers)
        response_data = json.loads(response.text, cls=RoundTripDecoder)
    except Exception as e:
        print("ERROR in call_cloud_function: {}".format(str(e)))

    return response_data


class MonkeeTracker:  # --------------------------------------------------------------------
    def __init__(self, db, app_name, function_name, human_uid):
        self.tracker = []
        self.db = db
        self.app_name = app_name
        self.human_uid = human_uid
        self.function_name = function_name

    def set_db(self, db_destination):
        self.db = db_destination

    def set_human_uid(self, human_uid):
        self.human_uid = human_uid

    def track(self, event_name):
        self.tracker.append(
            {"ts": datetime.now(timezone.utc), "e": event_name})

    def persist(self, logUid=None):
        self.track("persist_tracker")
        if len(self.tracker) >= 1:
            diff = self.tracker[-1]["ts"] - self.tracker[0]["ts"]
        overall_diff_s = diff.seconds + diff.microseconds / 1000000

        ti = 0
        while ti < len(self.tracker) - 1:
            next_time = self.tracker[ti + 1]["ts"]
            elapsed_mcs = next_time - self.tracker[ti]["ts"]
            self.tracker[ti]["elapsed_time"] = (
                elapsed_mcs.seconds + elapsed_mcs.microseconds / 1000000
            )
            self.tracker[ti]["a_perc"] = (
                str(
                    int(
                        round(
                            100 *
                            self.tracker[ti]["elapsed_time"] /
                            overall_diff_s, 0
                        )
                    )
                )
                + "%"
            )
            ti += 1

        track_dict = {
            "humanUid": self.human_uid,
            "calledFrom": self.function_name,
            "duration": overall_diff_s,
            "log": self.tracker,
        }
        r = int(10000 * random.random())
        seconds = 9999999999 - time.time()
        logUid = str(seconds) + str(r)
        if logUid is None:
            self.db.document().set(track_dict)
        else:
            self.db.document(logUid).set(track_dict)


def get_size(json_obj):
    """
    returns the size of the JSON object in bytes
    """
    dumps = json.dumps(json_obj, cls=RoundTripEncoder)
    size_bytes = sys.getsizeof(dumps)
    return size_bytes


def dateDiff(unit, ts1, ts2):
    """
    returns the time delta between ts1 and ts2 in the provided unit.
    Unit in: ['second','minute','hour','day']
    """
    elapsedTime = ts2 - ts1
    totalSeconds = elapsedTime.total_seconds()

    if unit in ["s", "sec", "second"]:
        return totalSeconds
    elif unit in ["mn", "min", "minute"]:
        return totalSeconds / 60
    elif unit in ["hr", "hour"]:
        return totalSeconds / 60 / 60
    elif unit in ["d", "day"]:
        return totalSeconds / 60 / 60 / 24


def getval(dictionary, key, default_value=None):
    if dictionary is not None:
        if key in dictionary:
            ret = dictionary[key]
        else:
            ret = default_value
    else:
        ret = default_value
    return ret


def get_uuid():
    return str(uuid.uuid4())


def describe_time(hh_ago):
    ret = ""
    hh_ago = int(round(hh_ago))
    if hh_ago == 0:
        ret = "very recently"
    elif hh_ago == 1:
        ret = "an hour ago"
    elif hh_ago <= 24:
        ret = str(hh_ago) + " hours ago"
    elif hh_ago <= 48:
        ret = "yesterday"
    else:
        dd_ago = int(round(hh_ago / 24))
        ret = str(dd_ago) + " days ago"

    return ret
